"""
FMH3 format information for pyfmh3
"""

_construct_version = None
try:
    from construct import __version__ as _construct_version
    _construct_version = [int(v) for v in _construct_version.split('.')]
except Exception:
    # I'd rather just continue than throw an error if this fails for some reason, like versioning changes,
    # so just let _construct_version be None
    # Users following instructions should never have a low version anyway
    pass

if _construct_version:
    if (_construct_version[0] < 2) or ((_construct_version[0] == 2) and (_construct_version[1] < 9)):
        raise Exception('Construct version too low, please install version 2.9+')

from construct import Struct, Computed, Tell, Const, Padding, Padded, Pointer, RepeatUntil, Byte, Int16ul, Flag, Rebuild, If, Int32ub, Seek, Int32ul, Int64ul
from pydiva.util.cs3_file_utils import RelocationPointerAdapter, gen_relocation_struct, relocation_data_len, gen_eofc_struct, gen_cs3_file

def _gen_fmh3_struct(int_type, pointer_type, addr_mode='rel'):
    return Struct(
        "pointer_offset" / (Computed(0) if addr_mode == 'abs' else Tell),
        "signature" / Const(b'FMH3'),
        Padding(4),
        "fonts_count" / Padded(pointer_type.sizeof(), int_type),
        "fonts_pointers_offset" / pointer_type,
        "fonts" / Pointer(lambda this: this.fonts_pointers_offset + this.pointer_offset, RepeatUntil(lambda obj,lst,ctx: ctx._index >= ctx.fonts_count - 1, Struct(
            "pointer" / pointer_type,
            "data" / Pointer(lambda this: this.pointer + this._.pointer_offset, Struct(
                "id" / int_type,
                "advance_width" / Byte,
                "line_height" / Byte,
                "box_width" / Byte,
                "box_height" / Byte,
                "layout_param_1" / Byte, 
                "layout_param_2_numerator" / Byte,
                "layout_param_2_denominator" / Byte,
                Padding(1),
                "other_params?" / int_type,
                "tex_size_chars" / int_type,
                "chars_count" / int_type,
                "chars_pointer" / pointer_type,
                "chars" / Pointer(lambda this: this.chars_pointer + this._._.pointer_offset, RepeatUntil(lambda obj,lst,ctx: ctx._index >= ctx.chars_count - 1, Struct(
                    "codepoint" / Int16ul,
                    "halfwidth" / Flag,
                    Padding(1),
                    "tex_col" / Byte,
                    "tex_row" / Byte,
                    "glyph_x" / Byte,
                    "glyph_width" / Byte,
                ))),
            )),
        ))),
    )

_fmh3_types = {
    'FMH3': {
        'remarks': 'unencapsulated FT fontmap',
        'struct': _gen_fmh3_struct(Int32ul, Int32ul),
        'address_size': 4,
        'fonts_pointers_min_offset': 32,
        'nest_fmh3_data': False,
    },
    'FONM': {
        'remarks': 'X fontmap in FONM container',
        'struct': gen_cs3_file(Int32ul, Int64ul, [{
            'signature': 'FONM',
            'data_size': lambda this: this.data_size,
            'data_subcon': _gen_fmh3_struct(Int32ul, RelocationPointerAdapter(Int64ul)),
            'enrs': True,
            'relocation': True
        }]),
        'address_size': 8,
        'fonts_pointers_min_offset': 32,
        'nest_fmh3_data': 'FONM',
        'alternate_type_checks': [{'type': 'FONM_F2', 'checks': [{'offset': 15, 'mask': 0x08}]}]
    },
    'FONM_F2': {
        'remarks': 'F2nd fontmap in FONM container',
        'struct': gen_cs3_file(Int32ub, Int32ub, [{
            'signature': 'FONM',
            'data_size': lambda this: this.data_size,
            'data_subcon': _gen_fmh3_struct(Int32ub, RelocationPointerAdapter(Int32ub), addr_mode='abs'),
            'enrs': False,
            'relocation': True
        }]),
        'address_size': 4,
        'fonts_pointers_min_offset': 32 + 64, # because FONM headers are within the same address space :/
        'nest_fmh3_data': 'FONM',
    },
}


"""
Stuff for dealing with sectioned from CS3 engine games. (F2nd, X)

Only single-section binary files (eg. FONM, DSC) + relocation are tested, so
other stuff is probably wrong, but the intent is that is can be expanded in the
future.
"""

from construct import Adapter, Struct, Tell, Const, Rebuild, Int32ub, Int32ul, Padding, Bytes, Default, If, Seek, Probe

def gen_relocation_data(offsets):
    """
    Given a list of offsets (offsets are in multiples of pointer length),
    output bytes to put in the relocation section's data.
    """
    
    last_offset = 0
    out = b''
    for o in offsets:
        distance = o - last_offset
        
        if distance > 0x3ff:
            b = distance.to_bytes(4, byteorder='big')
            b = bytes([b[0] | 0xc0]) + b[1:]
        elif distance > 0x3f:
            b = distance.to_bytes(2, byteorder='big')
            b = bytes([b[0] | 0x80]) + b[1:]
        else:
            b = distance.to_bytes(1, byteorder='big')
            b = bytes([b[0] | 0x40]) + b[1:]
        
        out += b
        last_offset = o
    
    out_size = len(out) + 4
    out = out_size.to_bytes(4, byteorder='little', signed=False) + out # prepend length
    
    # pad to 16 bytes
    while len(out) % 16:
        out += b'\x00'
    
    return out

def relocation_data_len(offsets):
    """
    Get the length of relocation data.
    (implemnted by calling len on gen_relocation_data result)
    """
    
    return len(gen_relocation_data(offsets))

class RelocationPointerAdapter(Adapter):
    """
    Wraps the pointer type, and generates relocation data during building.
    
    Relocation data will be put in root context's extra_sections['pof1']['data']
    (root context is up one level from level that contains pointer_offset)
    """
    
    def _decode(self, obj, context, path):
        # called at parsing time to return a modified version of obj
        
        return obj

    def _encode(self, obj, context, path):
        # called at building time to return a modified version of obj
        
        io_offset = context._io.tell()
        
        # try to find the current context's pointer offset by recursing up
        pointer_offset_ctx = context
        while hasattr(pointer_offset_ctx, '_') and not hasattr(pointer_offset_ctx, 'pointer_offset'):
            pointer_offset_ctx = pointer_offset_ctx._
        pointer_offset = getattr(pointer_offset_ctx, 'pointer_offset', 0)
        
        offset = io_offset - pointer_offset
        pointer_size = self.subcon.sizeof()
        if offset % pointer_size != 0:
            raise Exception('Pointer at bad offset -- can\'t generate relocation data')
        offset = offset // pointer_size
        
        # set offset to pointer_relocation_offsets in the parent context of pointer_offset
        # -- directly in the POF section's data (DIRTY SOLUTION!)
        # (doesn't make sense to not have this when using relocation tho)
        if not 'extra_sections' in pointer_offset_ctx._:
            pointer_offset_ctx._['extra_sections'] = {}
        if not 'POF' in pointer_offset_ctx._['extra_sections']:
            pointer_offset_ctx._['extra_sections']['POF'] = {}
        if not 'data' in pointer_offset_ctx._['extra_sections']['POF']:
            pointer_offset_ctx._['extra_sections']['POF']['data'] = []
        
        pointer_offset_ctx._['extra_sections']['POF']['data'] += [offset]
        
        pointer_offset_ctx._['extra_sections']['POF']['data'].sort()
        
        return obj

class RelocationDataAdapter(Adapter):
    """Converts a list of pointer offsets into relocation section data."""
    
    def _decode(self, obj, context, path):
        # called at parsing time to return a modified version of obj
        
        return obj

    def _encode(self, obj, context, path):
        # called at building time to return a modified version of obj
        
        return gen_relocation_data(obj)

def gen_relocation_struct(pointer_type, depth):
    """
    Generates a Construct struct for a POFx (relocation) section.
    
    Required context layout for data is:
    ```
    data=relocation_offsets
    ```
    """
    
    int_type = Int32ul # seems to use little endian in F2nd fontmap
    
    return Struct(
        "pointer_offset" / Tell,
        "signature" / Const(b'POF0' if pointer_type.sizeof() == 4 else b'POF1'),
        "section_size" / Rebuild(Int32ul, lambda this: relocation_data_len(this.data)),
        "data_pointer" / Default(Int32ul, 32),
        "flags" / Const(0x00000018 if int_type == Int32ub else 0x10000000, int_type),
        "depth" / Rebuild(Int32ul, depth),
        "data_size" / Rebuild(Int32ul, lambda this: this.section_size),
        Padding(lambda this: this.data_pointer - 24),
        Seek(lambda this: this.data_pointer + this.pointer_offset),
        "data" / RelocationDataAdapter(Bytes(lambda this: this.data_size))
    )

def gen_eofc_struct(depth, fix_parent_size=None):
    """Generates a Construct struct for an EOFC (end of file) section."""
    
    int_type = Int32ul # seems to use little endian in F2nd fontmap
    
    return Struct(
        "pointer_offset" / Tell,
        "signature" / Const(b'EOFC'),
        "section_size" / Rebuild(Int32ul, 0),
        "data_pointer" / Rebuild(Int32ul, 32),
        "flags" / Const(0x00000018 if int_type == Int32ub else 0x10000000, int_type),
        "depth" / Rebuild(Int32ul, depth),
        "data_size" / Rebuild(Int32ul, 0),
        Padding(lambda this: this.data_pointer - 24),
        Seek(lambda this: this.data_pointer + this.pointer_offset),
        Rebuild(Bytes(lambda this: this.data_size), b''),
        "parent_size_fix" / If(fix_parent_size, Struct(
            "end_offset" / Tell,
            Seek(lambda this: this._._.section_outer.section.section_size_offset),
            Rebuild(int_type, lambda this: this.end_offset - (this._._.section_outer.section.data_pointer + this._._.section_outer.section.pointer_offset)),
            Seek(lambda this: this.end_offset)
        ))
    )

def gen_section_struct(int_type, pointer_type, signature, depth, data_size=0, data_subcon=Bytes(lambda this: this.data_size), enrs=False, relocation=False, optional=False):
    """
    Generates a Construct scruct for an arbitrary section, with user-defined
    signature, data_size (can be lambda), and data subcon.
    enrs and relocation are optional (enrs only so parsing won't fail), and the
    entire section can be made optional.
    
    Optional sections are skipped if not present during parsing, and always
    skipped during building.
    
    To read data_size from section dict when building, you can use
    `lambda this: this.data_size`.
    
    Required context layout for data is:
    ```
    section_outer=dict(
        section=dict(
            data_pointer=32,
            data=b'Sample Data',
            extra_sections=dict() # needed to avoid KeyError during build with relocation
        )
    )
    ```
    """
    
    return "section_outer" / If(lambda this: this._parsing or not optional, Struct(
        "signature_first_pass" / Rebuild(Bytes(4), signature.encode('ascii')),
        Seek(lambda this: this._io.tell() - 4),
        "section" / If(lambda this: this.signature_first_pass == signature.encode('ascii') or not optional, Struct(
            "pointer_offset" / Tell,
            "signature" / Const(signature.encode('ascii')),
            "section_size_offset" / Tell, # for fixing section_size later 
            "section_size" / Default(Int32ul, 0),
            "data_pointer" / Default(Int32ul, 32),
            #Probe(),
            "flags" / Const(0x00000018 if int_type == Int32ub else 0x10000000, int_type),
            "depth" / Rebuild(Int32ul, depth),
            "data_size" / Rebuild(Int32ul, data_size),
            Padding(lambda this: this.data_pointer - 24),
            Seek(lambda this: this.data_pointer + this.pointer_offset),
            "data" / Default(data_subcon, b''),
            Seek(lambda this: this.data_pointer + this.data_size + this.pointer_offset),
            "extra_sections" / Struct(
                # I think enrs may come before or after relocation(?), so just try reading it from both
                gen_section_struct(int_type, pointer_type, 'ENRS', depth, optional=True) if enrs else Tell,
                "POF" / gen_relocation_struct(pointer_type, depth) if relocation else Tell,
                gen_section_struct(int_type, pointer_type, 'ENRS', depth, optional=True) if enrs else Tell,
                "end_offset" / Tell,
                Seek(lambda this: this._.section_size_offset),
                Rebuild(Int32ul, lambda this: this.end_offset - (this._.data_pointer + this._.pointer_offset)),
                Seek(lambda this: this.end_offset)
            ),
        ))
    ))

def gen_cs3_sections(int_type, pointer_type, sections, depth=0):
    """
    Generates a Construct scruct for multiple sections, including children,
    from a list of info dicts.
    
    My understanding is that these should be structured in the file like:
    [
        SECTION_MAIN
        SECTION_ENRS
        SECTION_POF
        SECTION_CHILDREN // may have nested complete sections
        SECTION_EOF
    ]
    but this is based only on skimming some MikuMikuLibrary code and I never checked it lol
    
    Info dicts must have:
    signature
    
    Info dicts may have:
    data_size, data_subcon (by default, will use 0 length and Bytes),
    enrs, relocation (set to enable sections - enrs only so parsing won't fail),
    optional (set to make this section optional to read, and disable building),
    child_sections (info list for this section's children)
    
    To read data_size from section dict when building, you can use
    `lambda this: this.data_size`.
    
    Required context layout for data is:
    (well, intended context...  this isn't very tested yet)
    ```
    '[SIGNATURE]': {
        section_outer=dict(
            section=dict(
                data_pointer=32,
                data=b'Sample Data',
                extra_sections=dict() # needed to avoid KeyError during build with relocation
            )
        ),
        '[CHILD0_SIG]'=dict(
            section_outer=dict(
                section=dict(
                    data_pointer=32,
                    data=b'Sample Data',
                    extra_sections=dict() # needed to avoid KeyError during build
                )
            ),
            '[CH0_CH1_SIG]'=dict(
                ...
            )
        )
        ...
    }
    ```
    """
    
    if len(sections) == 0:
        return Struct()
    
    struct = Struct()
    for s in sections:
        ss = gen_section_struct(int_type, pointer_type, s['signature'], depth, s.get('data_size', 0), s.get('data_subcon', Bytes(lambda this: this.data_size)), s.get('enrs'), s.get('relocation'), s.get('optional'))
        
        cs = gen_cs3_sections(int_type, pointer_type, s.get('child_sections', []), depth + 1)
        ss += cs
        
        ss += "EOFC" / gen_eofc_struct(depth, s['signature'])
        struct += s['signature'] / ss
    
    return struct

def gen_cs3_file(int_type, pointer_type, sections):
    """
    Generates a Construct scruct for a file with multiple sections, including
    children, from a list of info dicts.
    
    Same as gen_cs3_sections but with an end of file section appended.
    """
    
    return gen_cs3_sections(int_type, pointer_type, sections) + gen_eofc_struct(0)


"""
FT file helper functions for pyfarc
"""

from io import BytesIO
from secrets import token_bytes
from os import getenv

try:
    from Cryptodome.Cipher import AES
    from Cryptodome.Util.Padding import pad, unpad
except Exception:
    pass # main pyfarc module already handles this

def _needs_FT_decryption(s):
    """Checks if the FARC file stream needs FT-type decryption"""
    
    og_pos = s.tell()
    
    s.seek(og_pos + 11)
    encrypted = True if s.read(1)[0] & 4 else False
    
    s.seek(og_pos + 16)
    alignment = s.read(4)
    alignment_bits_popcnt = sum([bin(x).count('1') for x in alignment])
    
    s.seek(og_pos + 20)
    format = s.read(4)
    
    # heuristic-based detection similar to MML -- this should have a false negative rate of ~1 in 4.8b (compared to MML's ~1 in 134.2m)
    # false positives will require either more than 8 bits set in popcnt (extremely unlikely) or a new subformat which would be unsupported anyway
    # if many bits (>8) are set in alignment, it seems incorrect -- likely encrypted (~1 in 285 false negative)
    # if format doesn't start with null bytes, it seems incorrect -- likely encrypted (~1 in 16.8m false negative)
    # if the entire 16 bytes are null, it indicates a null IV and definite encryption (or a bad file)
    s.seek(og_pos + 16) # prepare for reading entire IV
    if encrypted and (alignment_bits_popcnt > 8 or format[:3] != b'\x00\x00\x00' or s.read(16) == b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'):
        s.seek(og_pos)
        return True
    
    s.seek(og_pos)
    return False

def _is_FT_FARC(s):
    """Returns whether the FARC file stream is in FT format"""
    
    og_pos = s.tell()
    magic_str = s.read(4).decode('ascii')
    s.seek(og_pos)
    
    if magic_str != 'FARC':
        return False
        
    if _needs_FT_decryption(s):
        return True
    
    
    s.seek(og_pos + 20)
    format = s.read(4)
    s.seek(og_pos)
    
    if format == b'\x00\x00\x00\x01':
        return True
    
    return False

def _decrypt_FT_FARC_header(s):
    """
    Decrypts header of FT FARC from stream and returns a new stream containing entire farc
    """
    
    if not _is_FT_FARC(s):
        return None
    
    if not _needs_FT_decryption(s):
        return None
    
    og_pos = s.tell()
    
    out = BytesIO()
    
    out.write(s.read(4)) # signature
    
    old_header_size = int.from_bytes(s.read(4), byteorder='big', signed=False)
    other_plaintext_header = s.read(8) # other plaintext header stuff
    
    cipher = AES.new(b'\x13\x72\xD5\x7B\x6E\x9E\x31\xEB\xA2\x39\xB8\x3C\x15\x57\xC6\xBB', AES.MODE_CBC, iv=s.read(16))
    header_data = cipher.decrypt(s.read(old_header_size - 16 - 8)) # 16 is for IV, 8 is for plaintext part of header
    header_data = unpad(header_data, 16, 'pkcs7')
    new_header_size = len(header_data) + 8
    
    out.write(new_header_size.to_bytes(4, byteorder='big', signed=False))
    out.write(other_plaintext_header)
    out.write(header_data)
    
    # resync streams (read old_header_size but wrote new_header_size)
    out.seek(out.tell() + old_header_size - new_header_size)
    
    out.write(s.read())
    
    s.seek(og_pos)
    out.seek(0)
    return out

def _encrypt_FT_FARC_header(instream, outstream):
    """
    Encrypts header of FT FARC from instream and writes entire FARC to outstream
    Ensure input FARC has enough space for IV and AES padding after the header
    """
    
    if not _is_FT_FARC(instream) or _needs_FT_decryption(instream):
        # file doesn't need encryption???
        #outstream.write(instream.read())
        #return
        raise Exception('Wrong format FARC or already encrypted')
    
    outstream.write(instream.read(4)) # signature
    
    old_header_size = int.from_bytes(instream.read(4), byteorder='big', signed=False)
    new_header_size = old_header_size - 8 # temporarily remove stuff that isn't encrypted
    new_header_size += 16 # iv seems to count towards header size
    new_header_size += 16 - (new_header_size % 16) # space for AES padding
    new_header_size += 8 # add size of plaintext parts back
    
    outstream.write(new_header_size.to_bytes(4, byteorder='big', signed=False))
    outstream.write(instream.read(8)) # other plaintext header stuff
    
    header_data = instream.read(old_header_size - 8) # header_size is 8 bytes longer than encrypted portion
    header_data = pad(header_data, 16, 'pkcs7')
    
    if new_header_size != len(header_data) + 16 + 8: # 16 is for IV, 8 is for plaintext part of header
        raise Exception('Header size calc is bugged! Please report this! actual: {}, expected: {}'.format(len(header_data) + 16 + 8, new_header_size))
    
    if getenv('PYFARC_NULL_IV'):
        iv = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    else:
        iv = token_bytes(16)
    outstream.write(iv)
    cipher = AES.new(b'\x13\x72\xD5\x7B\x6E\x9E\x31\xEB\xA2\x39\xB8\x3C\x15\x57\xC6\xBB', AES.MODE_CBC, iv=iv)
    outstream.write(cipher.encrypt(header_data))
    
    # resync streams (read old_header_size but wrote new_header_size)
    instream.seek(instream.tell() + new_header_size - old_header_size)
    
    outstream.write(instream.read())
    
    return


import setuptools

with open("readme.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pydiva",
    version="1.1.0",
    author="somewhatlurker",
    description="Some stuff for handling files from Project DIVA games",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/somewhatlurker/pydiva",
    packages=setuptools.find_packages(exclude=['tests*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
       'construct>=2.9.0',
       'pycryptodomex',
    ],
)


"""
pyfmh3 reader and writer for Project DIVA FMH3 fontmaps
can read+write AFT FMH3 and read X FONM
"""

from copy import deepcopy
from io import BytesIO
from pydiva.pyfmh3_formats import _fmh3_types


class UnsupportedFmh3TypeException(Exception):
    pass

def check_fmh3_type(t):
    """Checks if a fontmap type is supported and returns a remarks string. Raises UnsupportedFmh3TypeException if not supported."""
    
    if not t in _fmh3_types:
        raise UnsupportedFmh3TypeException("{} type not supported".format(t))
    
    return _fmh3_types[t]['remarks']


_fonts_pointers_min_offset = 32 # keep this a multiple of 8
_fonts_header_size_per_font = 32 # keep this a multiple of 8
_char_data_size = 8

def _fonts_header_pointers_size(n, address_size):
    """Returns the size of the pointers to font headers for the given number of fonts."""
    
    res = n * address_size
    if res % 16: res += 16 - (res % 16) # align to 16 bytes, keep this a multiple of 8
    return res

def _fonts_header_size(n):
    """Returns the size of the fonts header section for the given number of fonts."""
    
    return n * _fonts_header_size_per_font

def _char_array_size(n):
    """Returns the size of a character info array for n characters."""
    
    res = n * _char_data_size
    if res % 16: res += 16 - (res % 16) # keep this a multiple of 8
    return res

def _set_font_pointers(fonts, address_size):
    """Sets font header pointers in the given list of fonts"""
    
    pos = _fonts_pointers_min_offset + _fonts_header_pointers_size(len(fonts), address_size)
    
    for font in fonts:
        font['pointer'] = pos
        pos += _fonts_header_size_per_font

def _set_char_pointers(fonts, address_size):
    """Sets character info pointers in the given list of fonts"""
    
    pos = _fonts_pointers_min_offset + _fonts_header_pointers_size(len(fonts), address_size) + _fonts_header_size(len(fonts))
    
    for font in fonts:
        font['data']['chars_pointer'] = pos
        pos += _char_array_size(len(font['data']['chars']))

def _get_fmh3_length(fonts):
    """Gets the full length of the FMH3 data from a font with pointers and char counts already set"""
    
    return fonts[-1]['data']['chars_pointer'] + _char_array_size(fonts[-1]['data']['chars_count'])


def to_stream(data, stream, no_copy=False):
    """
    Converts a dictionary (formatted like the dictionary returned by from_stream) to fontmap data and writes it to a stream.
    
    Set no_copy to True for a speedup and memory usage reduction if you don't mind your input data being contaminated.
    """
    
    magic_str = data['fmh3_type']
    check_fmh3_type(magic_str)
    fmh3_type = _fmh3_types[magic_str]
    
    if no_copy:
        fonts = data['fonts']
    if not no_copy:
        fonts = deepcopy(data['fonts'])
    
    for font in fonts:
        font['chars_count'] = len(font['chars'])
    fonts = [{'data': font} for font in fonts]
    
    global _fonts_pointers_min_offset
    _fonts_pointers_min_offset = fmh3_type['fonts_pointers_min_offset']
    
    address_size = fmh3_type['address_size']
    _set_font_pointers(fonts, address_size)
    _set_char_pointers(fonts, address_size)
    
    if fmh3_type['nest_fmh3_data']:
        data_size =_get_fmh3_length(fonts)
        
        return fmh3_type['struct'].build_stream({
            fmh3_type['nest_fmh3_data']: dict(
                section_outer=dict(
                    section=dict(
                        data_pointer=64,
                        data_size=data_size,
                        data=dict(
                            fonts_count=len(fonts),
                            fonts_pointers_offset=_fonts_pointers_min_offset,
                            fonts=fonts
                        ),
                        extra_sections=dict() # needed to avoid KeyError during build
                    )
                )
            )
        }, stream)
    else:
        return fmh3_type['struct'].build_stream(dict(
            fonts_count=len(fonts),
            fonts_pointers_offset=_fonts_pointers_min_offset,
            fonts=fonts
        ), stream)

def to_bytes(data, no_copy=False):
    """
    Converts a dictionary (formatted like the dictionary returned by from_bytes) to an in-memory bytes object containing fontmap data.
    
    Set no_copy to True for a speedup and memory usage reduction if you don't mind your input data being contaminated.
    """
    
    with BytesIO() as s:
        to_stream(data, s, no_copy)
        return s.getvalue()


def _parsed_to_dict(fmhdata, nested_fmh_section):
    """Converts the raw construct data to our standard dictionary format."""
    
    if nested_fmh_section:
        magic_str = fmhdata[nested_fmh_section]['section_outer']['section']['signature'].decode('ascii')
        fmhdata = fmhdata[nested_fmh_section]['section_outer']['section']['data']
    else:
        magic_str = fmhdata['signature'].decode('ascii')
    
    fonts = []
    
    for font in fmhdata['fonts']:
        tmp = dict(font['data'])
        del tmp['_io']
        del tmp['chars_count']
        del tmp['chars_pointer']
        tmp['chars'] = [dict(char) for char in tmp['chars']]
        for char in tmp['chars']:
            del char['_io']
        fonts += [tmp]
    
    return {'fmh3_type': magic_str, 'fonts': fonts}

def from_stream(s):
    """Converts fontmap data from a stream to a dictionary."""
    
    pos = s.tell()
    magic_str = s.read(4).decode('ascii')
    s.seek(pos)
    
    check_fmh3_type(magic_str)
    fmh3_type = _fmh3_types[magic_str]
    
    if 'alternate_type_checks' in fmh3_type:
        for t in fmh3_type['alternate_type_checks']:
            match = True
            for c in t['checks']:
                s.seek(pos+c['offset'])
                if not int.from_bytes(s.read(1), byteorder='little') & c['mask']:
                    match = False
                    break
            
            if match: # passed all checks, to this is actually the alt type
                magic_str = t['type']
                fmh3_type = _fmh3_types[magic_str]
                break
    s.seek(pos)
    
    fmhdata = fmh3_type['struct'].parse_stream(s)
    res = _parsed_to_dict(fmhdata, fmh3_type['nest_fmh3_data'])
    res['fmh3_type'] = magic_str # force type to what we already determined
    return res

def from_bytes(b):
    """Converts fontmap data from bytes to a dictionary."""
    
    with BytesIO(b) as s:
        return from_stream(s)


# test_fmh = {'fmh3_type': 'FMH3', 'fonts': [{"id":2, "advance_width":24, "line_height":30, "box_width":26, "box_height":32, "layout_param_1":3, "layout_param_2_numerator":1, "layout_param_2_denominator":1, "other_params?":0, "tex_size_chars":19, "chars":[{"codepoint":48, "halfwidth":False, "tex_col":0, "tex_row":0, "glyph_x":0, "glyph_width":24}, {"codepoint":49, "halfwidth":False, "tex_col":1, "tex_row":0, "glyph_x":0, "glyph_width":24}]}]}